//---------------------------------------------------------------------------

#pragma hdrstop

#define _EXPORTING
#include "..\miniscope\device\MiniscopeDevice.h"
#include "..\miniscope\device\MiniscopeDeviceCapabilities.h"
#include "SignalCfg.h"
#include "Signal.h"
#include <assert.h>
#include "TestScope.h"
#include "Signal.h"
#include <algorithm>	// needed by Utils::in_group
#include "Utils.h"
#include <string>
#include <fstream> 
#include <json/json.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)

extern Signal signal[CHANNELS];

static const struct S_DEVICE_DLL_INTERFACE dll_interface =
	{DLL_INTERFACE_MAJOR_VERSION, DLL_INTERFACE_MINOR_VERSION};

// callback ptrs
CALLBACK_LOG lpLogFn = NULL;
CALLBACK_CONNECT lpConnectFn = NULL;
CALLBACK_DATA lpDataFn = NULL;
CALLBACK_TRIGGER lpTriggerFn = NULL;

void *callbackCookie;	///< used by upper class to distinguish library instances when receiving callbacks

void __stdcall GetMiniscopeInterfaceDescription(struct S_DEVICE_DLL_INTERFACE* interf)
{
	interf->majorVersion = dll_interface.majorVersion;
	interf->minorVersion = dll_interface.minorVersion;
}

void Log(char* txt)
{
	if (lpLogFn)
		lpLogFn(callbackCookie, txt);
}

void Data(short* data, int size)
{
	if (lpDataFn)
		lpDataFn(callbackCookie, data, size);
}

void Connect(int state, char *szMsgText)
{
	if (lpConnectFn)
		lpConnectFn(callbackCookie, state, szMsgText);
}

void Trigger(void)
{
	if (lpTriggerFn)
		lpTriggerFn(callbackCookie);
}

void __stdcall SetCallbacks(void *cookie, CALLBACK_LOG lpLog, CALLBACK_CONNECT lpConnect,
	CALLBACK_DATA lpData, CALLBACK_TRIGGER lpTrigger)
{
	assert(cookie && lpLog && lpConnect && lpData && lpTrigger);
	lpLogFn = lpLog;
	lpConnectFn = lpConnect;
	lpDataFn = lpData;
	lpTriggerFn = lpTrigger;
	callbackCookie = cookie;
	lpLogFn(callbackCookie, "sig_test.dll loaded. Yay!");

	testScope.callbackLog = Log;
	testScope.callbackData = Data;
	testScope.callbackConnect = Connect;
	testScope.callbackTrigger = Trigger;
}

static enum E_TRIGGER_TYPE eTrigger[] = { TRIGGER_MANUAL, TRIGGER_SLOPE_UP, TRIGGER_SLOPE_DOWN, TRIGGER_CONTINUOUS };
static enum E_COUPLING_TYPE eCoupling[] = { COUPLING_DC, COUPLING_AC };
unsigned int iBuf[] = { 512, 1024, 2048, 4096, 8192, 14336 };
float fSens[] = { 0.000547, 0.001602, 0.004766, 0.014335, 0.042969, 0.128906 };
float fSampl[] = { 2.048e-3, 1.024e-3, 5.12e-4, 2.56e-4, 1.28e-4, 6.4e-5,
	3.2e-5, 1.6e-5, 8e-6, 4e-6, 2e-6 };
static struct S_SCOPE_CAPABILITIES capabilities =
{
	CHANNELS,
	8,		//bits/sample
	sizeof(fSens)/sizeof(fSens[0]), fSens,
	sizeof(fSampl)/sizeof(fSampl[0]), fSampl,
	sizeof(eTrigger)/sizeof(eTrigger[0]), eTrigger,
	sizeof(eCoupling)/sizeof(eCoupling[0]), eCoupling,
	sizeof(iBuf)/sizeof(iBuf[0]), iBuf
};

unsigned int pSensitivity[CHANNELS];
enum E_COUPLING_TYPE eCouplingType[CHANNELS];

extern Signal signal[CHANNELS];

/** \brief Calibration data for single sensitivity range
*/
static struct S_CALIBRATION
{
	int offset;
	float gain;
} calibration[CHANNELS][sizeof(fSens)/sizeof(fSens[0])];

void __stdcall GetDeviceCapabilities(struct S_SCOPE_CAPABILITIES **caps)
{
	*caps = &capabilities;
}

void __stdcall ShowSettings(HANDLE parent)
{
	frmSignalConfig->Show();
}

int __stdcall SetSensitivity(int iChannelId, int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(fSens)/sizeof(fSens[0])))
		return -1;
	return testScope.SetSensitivity(iChannelId, iId);
}

int __stdcall SetSamplingPeriod(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(fSampl)/sizeof(fSampl[0])))
		return -1;
	return testScope.SetSampling(iId);
}

int __stdcall SetTriggerSource(int channel, int iId)
{
	if (channel < 0 || channel >= CHANNELS)
		return -1;
	if (iId < 0 || iId >= (int)(sizeof(eTrigger)/sizeof(enum E_TRIGGER_TYPE)))
		return -2;
	return testScope.SetTriggerType(channel, eTrigger[iId]);
}

int __stdcall SetTriggerLevel(int level)
{
	return testScope.SetTriggerLevel(level);	
}

int __stdcall SetCouplingType(int iChannel, int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(eCoupling)/sizeof(eCoupling[0])))
		return -1;
	return testScope.SetCouplingType(iChannel, eCoupling[iId]);
}

int __stdcall SetBufferSize(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(iBuf)/sizeof(unsigned int)))
		return -1;
	return testScope.SetBufferSize(iBuf[iId]);
}

int __stdcall Connect(void)
{
	if (frmSignalConfig == NULL)
	{
    	frmSignalConfig = new TfrmSignalConfig(NULL, signal, CHANNELS);
	}
	return testScope.Connect();
}

int __stdcall Disconnect(void)
{
	if (frmSignalConfig)
	{
		delete frmSignalConfig;
		frmSignalConfig = NULL;
	}
	return testScope.Disconnect();
}

int __stdcall AcquireStart(void)
{
	testScope.SetTriggerType(0, TRIGGER_CONTINUOUS);
	return testScope.ManualTrigger();
}

int __stdcall AcquireStop(void)
{
	return testScope.SetTriggerType(0, TRIGGER_MANUAL);
}

int __stdcall ManualTrigger(void)
{
	return testScope.ManualTrigger();
}

static bool bSettingsReaded = false;

static int GetDefaultSettings(struct S_SCOPE_SETTINGS* settings)
{
	for (unsigned int i=0; i<capabilities.iChannelCount; i++)
	{
		settings->pSensitivity[i] = 0;
		settings->eCouplingType[i] = COUPLING_DC;
	}

	settings->iSamplingPeriod = 0;
	settings->iTriggerSrcChannel = 0;
	settings->eTriggerType = TRIGGER_MANUAL;
	settings->iTriggerLevel = 0;
	settings->iBufferSize = 512;
	for (int i=0; i<CHANNELS; i++)
	{
		for (unsigned int j=0; j<sizeof(calibration)/sizeof(calibration[0]); j++)
		{
			calibration[i][j].offset = 0;
			calibration[i][j].gain = 1.0;
		}
	}
	bSettingsReaded = true;
	return 0;
}

int __stdcall GetMiniscopeSettings(struct S_SCOPE_SETTINGS* settings)
{
	std::string path = Utils::GetDllPath();
	path = Utils::ReplaceFileExtension(path, ".cfg");
	if (path == "")
		return GetDefaultSettings(settings);

	Json::Value root;   // will contains the root value after parsing.
	Json::Reader reader;

	std::ifstream ifs(path.c_str());
	std::string strConfig((std::istreambuf_iterator<char>(ifs)), std::istreambuf_iterator<char>());
	ifs.close();

	settings->pSensitivity = pSensitivity;
	settings->eCouplingType = eCouplingType;

	bool parsingSuccessful = reader.parse( strConfig, root );
	if ( !parsingSuccessful )
		return GetDefaultSettings(settings);

	for (unsigned int i=0; i<capabilities.iChannelCount; i++)
	{
		settings->pSensitivity[i] = root["Sensitivity"][i].asInt();
		if (settings->pSensitivity[i] >= sizeof(fSens)/sizeof(fSens[0]))
			settings->pSensitivity[i] = 0;

		int coupling = root["CouplingType"][i].asInt();
		if (Utils::in_group(coupling, eCoupling))
			settings->eCouplingType[i] = (E_COUPLING_TYPE)coupling;
		else
			settings->eCouplingType[i] = COUPLING_DC;
	}

	settings->iSamplingPeriod = root.get("SamplingPeriod", 0).asInt();
	if (settings->iSamplingPeriod >= sizeof(fSampl)/sizeof(fSampl[0]))
		settings->iSamplingPeriod = 0;

	int trigger = root.get("TriggerType", TRIGGER_MANUAL).asInt();
	if (Utils::in_group(trigger, eTrigger))
		settings->eTriggerType = (E_TRIGGER_TYPE)trigger;
	else
		settings->eTriggerType = TRIGGER_MANUAL;

	settings->iTriggerSrcChannel = 0;
	
	settings->iTriggerLevel = root.get("TriggerLevel", 0).asInt();
	if (settings->iTriggerLevel < -100 || settings->iTriggerLevel > 100)
		settings->iTriggerLevel = 0;

	settings->iBufferSize = root.get("BufferSize", 512).asInt();
	if (!Utils::in_group(settings->iBufferSize, iBuf))
		settings->iBufferSize = iBuf[0];

	// read calibration data also
	const Json::Value rootcal = root["Calibration"];	
	for (int i=0; i<CHANNELS; i++)
	{
		const Json::Value ch_cal = rootcal[i];
		if (ch_cal.type() != Json::arrayValue)
		{
			for (unsigned int j=0; j<sizeof(calibration[i])/sizeof(calibration[i][0]); j++)
			{
				calibration[i][j].offset = 0;
				calibration[i][j].gain = 1.0f;
			}
		}
		else
		{
			for (unsigned int j=0; j<sizeof(calibration[i])/sizeof(calibration[i][0]); j++)
			{
				calibration[i][j].offset = rootcal[i][j].get("Offset", 0).asInt();
				calibration[i][j].gain = rootcal[i][j].get("Gain", 1.0).asDouble();
			}
		}
	}

	// read signal settings
	const Json::Value rootsig = root["Signal"];
	if (rootsig.type() == Json::arrayValue)
	{
		for (unsigned int i=0; i<std::min((unsigned int)CHANNELS, rootsig.size()); i++)
		{
			const Json::Value sig = rootsig[i];
			if (sig.type() == Json::arrayValue)
			{
				for (unsigned int j=0; j<std::min((unsigned int)(Signal::SHAPE_CNT), sig.size()); j++)
				{
					signal[i].shape[j].shape = (Signal::E_WAVEFORM)sig[j].get("Shape", Signal::WAVE_NONE).asInt();
					signal[i].shape[j].fAmplitude = sig[j].get("Amplitude", 0.0f).asDouble();
					signal[i].shape[j].fFrequency = sig[j].get("Frequency", 0.0f).asDouble();
					signal[i].shape[j].fPhase = sig[j].get("Phase", 0.0f).asDouble();
				}
			}
		}
	}

	bSettingsReaded = true;
	return 0;
}

int __stdcall SaveMiniscopeSettings(struct S_SCOPE_SETTINGS* settings)
{
    Json::Value root;
	Json::StyledWriter writer;
	for (unsigned int i=0; i<capabilities.iChannelCount; i++)
	{
		root["Sensitivity"][i] = settings->pSensitivity[i];
		root["CouplingType"][i] = settings->eCouplingType[i];
	}
	root["SamplingPeriod"] = settings->iSamplingPeriod;
	root["TriggerType"] = settings->eTriggerType;
	root["TriggerLevel"] = settings->iTriggerLevel;
	root["BufferSize"] = settings->iBufferSize;

	// add calibration data also
	const Json::Value rootcal = root["Calibration"];
	for (unsigned int i=0; i<capabilities.iChannelCount; i++)
	{
		for (unsigned int j=0; j<sizeof(calibration)/sizeof(calibration[0]); j++)
		{
			  root["Calibration"][i][j]["Offset"] = calibration[i][j].offset;
			  root["Calibration"][i][j]["Gain"] = calibration[i][j].gain;
		}
	}

	// write signal settings
	for (unsigned int i=0; i<CHANNELS; i++)
	{
		for (unsigned int j=0; j<signal[i].shape.size(); j++)
		{
			root["Signal"][i][j]["Shape"] = signal[i].shape[j].shape;
			root["Signal"][i][j]["Amplitude"] = signal[i].shape[j].fAmplitude;
			root["Signal"][i][j]["Frequency"] = signal[i].shape[j].fFrequency;
			root["Signal"][i][j]["Phase"] = signal[i].shape[j].fPhase;
		}
	}

	std::string outputConfig = writer.write( root );

	std::string path = Utils::GetDllPath();
	path = Utils::ReplaceFileExtension(path, ".cfg");
	if (path == "")
		return -1;

    std::ofstream ofs(path.c_str());
    ofs << outputConfig;
	ofs.close();

	return 0;
}

int __stdcall StoreCalibration(int channel, int range, int offset, float gain)
{
	if (channel < 0 || channel >= CHANNELS)
		return -1;
	if (range < 0 || range >= (int)(sizeof(fSens)/sizeof(fSens[0])))
		return -2;
	calibration[channel][range].offset = offset;
	calibration[channel][range].gain = gain;
	return 0;
}

int __stdcall GetCalibration(int channel, int range, int* const offset, float* const gain)
{
	if (channel < 0 || channel >= CHANNELS)
		return -1;
	if (range < 0 || range >= (int)(sizeof(fSens)/sizeof(fSens[0])))
		return -2;
	struct S_SCOPE_SETTINGS settings;
	if (bSettingsReaded == false)
	{
		int status = GetMiniscopeSettings(&settings);
		if (status)
			return status;
	}
	*offset = calibration[channel][range].offset;
	*gain = calibration[channel][range].gain;
	return 0;
}

